/* 1200, Wed 3 Feb 99

   FD_PARSE.C:  Scanner/parser for fd_filter and fd_extract

   Copyright (C) 1992-2006 by Nevil Brownlee,
   CAIDA | University of Auckland */

/*
 * $Log: fd_parse.c,v $
 * Revision 1.1.1.2  2006/03/21 18:08:48  nevil
 * Version 5.1b15
 *
 * Revision 1.1.1.2.2.16  2002/10/31 23:02:26  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * Revision 1.1.1.2.2.15  2002/09/13 15:05:12  nevil
 * Implement VLANid, Priority and ECNCodeBits attributes
 *
 * Revision 1.1.1.2.2.14  2002/08/27 17:13:07  nevil
 * Improve thread locking.  Implement STREAM_TO_FLOW, ECNBits and TCPStreamData
 *
 * Revision 1.1.1.2.2.13  2002/06/29 01:55:41  nevil
 * Miscellaneous fixes for warnings found by Compaq Unix cc
 *
 * Revision 1.1.1.2.2.6  2000/08/08 19:44:42  nevil
 * 44b8 release
 *
 * Revision 1.1.1.2.2.4  2000/06/06 03:38:08  nevil
 * Combine NEW_ATR with TCP_ATR, various bug fixes
 *
 * Revision 1.1.1.2.2.1  2000/01/12 02:57:03  nevil
 * Implement 'packet pair matched' turnaroundtime distribution attributes.
 * Fix ASN-related bugs in NeTraMet, distribution-related bugs in fd_filter.
 *
 * Revision 1.1.1.2  1999/10/03 21:06:14  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.9  1999/09/22 05:38:34  nevil
 * Improve code to work properly on 64-bit machines
 * - Add OS=ALPHA handling to configure.in
 * - Clean up the Alpha compiler warnings
 * - Change all the snmp-related code to use Bit32 instead of unsigned long
 *
 * Revision 1.1.1.1.2.8  1999/09/14 00:46:46  nevil
 * 4.3 Release ..
 *  - Implement -D option to run NeMaC as a daemon
 *  - Tidy up the on-line help displays
 *  - Make IPv4 the default PeerType for sfmt_attribute()
 *
 * Revision 1.1.1.1.2.7  1999/05/18 03:36:24  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.6  1999/02/15 21:24:05  nevil
 * Distribution file for 4.3b9
 *
 * Revision 1.1.1.1.2.5  1999/02/03 04:41:42  nevil
 * Implementation of TCP attributes, part 5
 *
 * Revision 1.1.1.1.2.4  1999/01/27 04:26:12  nevil
 * Minor corrections to fix compiler warnings
 *
 * Revision 1.1.1.1.2.3  1999/01/20 04:01:32  nevil
 * Implementation of TCP attributes, part 4
 *
 * Revision 1.1.1.1.2.2  1999/01/08 01:37:07  nevil
 * Implement TCPdata and DScodepoint attribute
 *
 * Revision 1.1.1.1.2.1  1998/11/26 04:13:49  nevil
 * Fix problems in fd_filter and fd_extract when dealing with 'new'
 * attributes, especially those used with NETFLOW. In particular:
 * - Make sure attribs[] is referred to via attr_ix[]
 * - Get the separators correctly from a #Format record (no quotes)
 * - fd_extract wasn't writing out its data for the last sample!
 *
 * Revision 1.1.1.1  1998/11/16 03:57:27  nevil
 * Import of NeTraMet 4.3b3
 *
 * Revision 1.1.1.1  1998/11/16 03:22:00  nevil
 * Import of release 4.3b3
 *
 * Revision 1.1.1.1.2.1  1998/11/11 23:14:36  nevil
 * Only include malloc.h if we HAVE_MALLOC_H
 *
 * Revision 1.1.1.1  1998/10/28 20:31:23  nevil
 * Import of NeTraMet 4.3b1
 *
 * Revision 1.1.3.2  1998/10/18 23:44:04  nevil
 * Added Nicolai's patches, some 'tidying up' of the source
 *
 * Revision 1.1.3.1  1998/10/13 02:48:17  nevil
 * Import of Nicolai's 4.2.2
 *
 * Revision 1.1.1.1  1998/08/24 12:09:28  nguba
 * NetraMet 4.2 Original Distribution
 *
 * Revision 1.4  1998/07/21 00:43:57  rtfm
 * Change attrib numbers for 'New Attribs' I-D
 * First release version of SRL
 *
 * Revision 1.3  1998/05/07 04:28:53  rtfm
 * Implement NetFlowMet, the Cisco NetFlow RTFM meter
 */

#if HAVE_CONFIG_H
#include <ntm_conf.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <time.h>
#include <netinet/in.h>

#include "ausnmp.h"
#include "asn1.h"

#include "fd_data.h"
#include "nmc_c64.h"

void quack(int n) {}  /* For gdb */

#if NEW_ATR
#define TEST_DELTA      0
#define TEST_TRANSFORM  0
#define VALUE_TRACE     0

struct distribution *dl INIT(NULL);

struct distribution *get_dist(void)  /* Get an unused distribution */
{
   struct distribution *d;
   if (dl != NULL) {
      d = dl;  dl = dl->next;
      }
   else {
      d = (struct distribution *)malloc(sizeof(struct distribution));
      }
   return d;
   }

void free_dist(struct distribution *d)  /* Return distribution to free list */
{
   d->next = dl;
   dl = d;
   }
#endif  /* NEW_ATR */

char *gnbr(unsigned int *n,char *s)  /* Get nbr from 's', */
{
   unsigned int v = 0, b = 10, d;
   while (*s == ' ') ++s;
   if (*s == 'x' || *s == 'X') {
      ++s;  b = 16;
      }
   else if (*s == '0') {
      ++s;  b = 8;
      }
   for (;;) {
      d = *s;
      if (b == 16) {
         if (!isxdigit(d)) break;
         }
      else {
         if (!isdigit(d)) break;
         if (b == 8 && d > '7') break;
         }
      if (d <= '9') d -= '0';
      else if (d <= 'F') d -= ('A'-10);
      else d -= ('a'-10);
      v = v*b + d;  ++s;
      }
   *n = v;  return s;
   }

char *gcstring(char *s, int *len)  /* Get string from s */
{
   static char escin[] = {
      'b', 'f', 'n', 'r', 't', 'v','\\','\'','\"','\?', 0 };
   static char escout[] ={
      '\b','\f','\n','\r','\t','\v','\\','\'','\"','\?' };
   unsigned int c,j;
   char sbuf[80], *t = sbuf, *rp;
   while (*s) {
      if (*s == '\\') {
         ++s;  c = *s;
         if (c == '0' || c == 'x' || c == 'X') {  /* Octal or hex nbr */
            s = gnbr(&c,s);  *t++ = c;
	    }
         else if (isdigit(c)) {  /* Octal number */
            --s;  *s = '0';  /* Make gnbr use base 8 */
            s = gnbr(&c,s);  *t++ = c;
	    }
         else {
            j = 0;  do {
               if (c == escin[j]) break;
	       } while (escin[++j] != 0);
            if (escin[j] != 0) {
               *t++ = escout[j];  ++s;
	       }
	    }
         }
      else *t++ = *s++;
      }
   *t = '\0';  *len = t-sbuf;
   rp = (char *)malloc(*len+1);  strcpy(rp,sbuf);
   return rp;
   }

int parse_open(char *fn)  /* Open file and initialise parser */
{
   if ((rfp = fopen(fn, "r")) == NULL) return 0;
   rferrors = rule_line = 0;  ic = '\n';
   return 1;
   }

void p_error(char *msg)
{
   if (!iblisted) {
      fprintf(stderr,"line %d: %s\n", rule_line,inbuf);
      iblisted = 1;
      }
   fprintf(stderr,"%s !!!\n", msg);
   ++rferrors;
   }

int nextchar(void)
{
   char *p;
   lic = ic;
   for (;;) {
      if (lic == '\n') {
	 if (fgets(inbuf, sizeof(inbuf), rfp) == NULL) return ic = EOF;
	 iblisted = 0;  ++rule_line;
	 ibp = inbuf;
	 }
      ic = *ibp++;
      if (ic == '#') lic = '\n';  /* Ignore comments */
      else return ic;
      }
   }

int nexttoken(void)
{
   for (;;) {
      if (isalnum(ic) || ic == '_') return 1;
      if (ic == ';' || ic == EOF) return 0;
      nextchar();
      }
   }

int relop_next(void)
{
   int op;
   for (;;) {
      if (isalnum(ic) || ic == '_' || ic == ';' || ic == EOF)
         return 0;
      if (ic == '=' || ic == '!') {
         op = ic;  nextchar();
         switch (op) {
         case '=':
            return OP_EQ;
         case '!':
            return OP_NE;
            }
         }
      nextchar();
      }
   }

void getfdstring(unsigned char *arg)
{
   for (;;) {
      if (isxdigit(ic)) break;  /* Addresses start with a (hex) digit */
#if V6
      if (ic == ':') break;  /* v6 addresses may start with a colon */
#endif
      nextchar();
      }
   if (ic != EOF) do {
      *arg++ = toupper(ic);
      nextchar();  /* Addresses may include . (decimal) or - (hex) */
      } while (isxdigit(ic) || ic == '.' || ic == '-'
#if V6
         || ic == ':'
#endif
         );
   *arg = '\0';  return;

   }

int wordis(char *p, char *w)
{
   if (strlen(p) != strlen(w)) return 0;  /* Lengths differ */
   return strcmp(p,w) == 0;
   }

char wbuf[50];  /* Remember the last word we found */

int getword(int res_reqd)
{
   char *wp = wbuf;
   int j;
   for (;;) {
      *wp++ = tolower(ic);
      nextchar();
      if (ic == EOF) return EOF;
      if (!isalnum(ic) && ic != '_') break;
      }
   *wp = '\0';

   if (wordis(wbuf,"set")) return RF_SET;
   if (wordis(wbuf,"format")) return RF_FORMAT;
   if (wordis(wbuf,"tag")) return RF_TAG;
   if (wordis(wbuf,"statistics")) return RF_STATS;
   if (wordis(wbuf,"column")) return RF_COLUMN;
   if (wordis(wbuf,"time")) return RF_TIME;

   if (wordis(wbuf,"scale")) return EX_SCALE;
   if (wordis(wbuf,"kilo")) return EX_KILO;
   if (wordis(wbuf,"mega")) return EX_MEGA;
   if (wordis(wbuf,"elapsed")) return EX_ELAPSED;
   if (wordis(wbuf,"clock")) return EX_CLOCK;
   if (wordis(wbuf,"seconds")) return EX_SECS;
   if (wordis(wbuf,"minutes")) return EX_MINS;
   if (wordis(wbuf,"hours")) return EX_HOURS;
   if (wordis(wbuf,"days")) return EX_DAYS;
   if (wordis(wbuf,"string")) return EX_TSTRING;

   if (wordis(wbuf,"ip")) return AT_IP4;
   if (wordis(wbuf,"clns")) return AT_CLNS;
   if (wordis(wbuf,"decnet")) return AT_DECNET;
   if (wordis(wbuf,"ipx")) return AT_NOVELL;
   if (wordis(wbuf,"novell")) return AT_NOVELL;
   if (wordis(wbuf,"ethertalk")) return AT_ETHERTALK;
   if (wordis(wbuf,"other")) return AT_OTHER;

   if (wordis(wbuf,"icmp")) return PT_ICMP;
   if (wordis(wbuf,"tcp")) return PT_TCP;
   if (wordis(wbuf,"udp")) return PT_UDP;
   if (wordis(wbuf,"ospf")) return PT_OSPF;

   if (wordis(wbuf,"ftpdata")) return WNP_FTPDATA;
   if (wordis(wbuf,"ftp")) return WNP_FTP;
   if (wordis(wbuf,"telnet")) return WNP_TELNET;
   if (wordis(wbuf,"smtp")) return WNP_SMTP;
   if (wordis(wbuf,"domain")) return WNP_DOMAIN;
   if (wordis(wbuf,"nntp")) return WNP_NNTP;
   if (wordis(wbuf,"ntp")) return WNP_NTP;
   if (wordis(wbuf,"snmp")) return WNP_SNMP;

   for (j = 0; j != SZ_ATTRIBS; ++j) {
      if (wordis(wbuf,attribs[j].name)) return attribs[j].index;
      }
   if (res_reqd) {
      sprintf(ebuf, "Unknown word %s", wbuf);
      p_error(ebuf);
      }
   return 0;
   }



int getnbr(void)  /* Words return negative numbers */
{
   int v = 0;
   for (;;) {
      if (ic == EOF) return EOF;
      if (isdigit(ic)) break;
      if (isalpha(ic)) return getword(1);
      if (ic == '\'') {
         v = nextchar();
         if (nextchar() != '\'')
            p_error("Missing closing quote");
         return v;
         }
      nextchar();
      }
   for (;;) {
      v = v*10 + ic-'0';
      if (nextchar() == EOF) return EOF;
      if (!isdigit(ic)) break;
      }
   return v;
   }

void scan_format(struct flow_stats *fs, int single_record, int list)
{
   char sbuf[32], *sp;
   int a, n, nt, len, x, nb;

   FILE *f = fs->flows;  int MIBver = fs->MIBver;
   int zfile = fs->zfile;
#if HAVE_LIBZ
   gzFile gz_flows = fs->gz_flows;
#endif
   char *filename = fs->filename;

   memset (fs, 0, sizeof(struct flow_stats));
   fs->flows = f;  fs->MIBver = MIBver;
   fs->zfile = zfile;  
#if HAVE_LIBZ
   fs->gz_flows = gz_flows;
#endif
   fs->filename = filename;

   n = getnbr();     
   for (a = 0; ; ++a) {
      if (n < 0 || n > LASTATTRIB || attr_ix[n] == 0) {
         sprintf(ebuf, "Attribute %d not allowed in format",n);
         p_error(ebuf);
         }
      else {
         fs->format[a] = n;  fs->required[n] = 1;
#if NEW_ATR
         if (DISTRIB_ATTRIB(n))
            fs->distrib_bits |= 1 << n-DISTRIB_BASE;
         else if (DISTRIB2_ATTRIB(n))
            fs->distrib2_bits |= 1 << n-DISTRIB2_BASE;
         else if (DELTA_DISTRIB(n))
            fs->delta_bits |= 1 << n-DELTA_DISTRIB_BASE;
         else if (DELTA_DISTRIB2(n))
            fs->delta2_bits |= 1 << n-DELTA_DISTRIB2_BASE;
#endif
         }
      if (single_record) {  /* #Format: record */
         for (sp = sbuf, nt = 0; ; ) {
            if (ic == '\n') break;
            else if (isalnum(ic)) {
               n = getword(0);
	       nt = n >= 0 && n <= LASTATTRIB && attr_ix[n] != 0;
	       if (nt) break;  /* Found an attribute */
               else sp = strmov(sp,wbuf);
               }
            else {
               *sp++ = ic;  nextchar();
	       }
            }
         *sp = '\0';
	 fs->separator[a] = (char *)malloc(strlen(sbuf)+1);
         strcpy(fs->separator[a],sbuf);
         if (ic == '\n' && !nt) break;  /* End of #Format record */
         if (!nt) n = getword(0);
         }
      else {  /* Reading format statement from filter file */
         fs->separator[a] = " ";  /* Default */
         while (isspace(ic)) nextchar();  /* Skip blanks */

         if (ic == ';') break;  /* End of format statement */

         if (!isalnum(ic)) {  /* Separator */
            if (ic == '\"') {  /* Separator string */
               nextchar();
               for (sp = sbuf; ; ) {
                  if (ic == '\"') break;
                  *sp++ = ic;
                  nextchar();
                  }
               nextchar();
               }
            else {
               sp = sbuf;  *sp++ = ' ';
               for (;;) {
                  if (isspace(ic)) break;
                  *sp++ = ic;
	          nextchar();
                  }
               *sp++ = ' ';
	       }
            *sp = '\0';
	    fs->separator[a] = gcstring(sbuf, &len);
            while (isspace(ic)) nextchar();  /* Skip blanks */
            }

         if (ic == ';') break;  /* End of format statement */
         n = getnbr();     
         }
      }

   fs->n_attr = a+1;
   if (list && !quiet) {
      fprintf(stderr,"Format: ");
      for (n = 0; n != fs->n_attr; ++n) {
         a = fs->format[n];
         x = attr_ix[a];  /* Index into attribs[] */
         fprintf(stderr,attribs[x].name);
         if (n != fs->n_attr-1
               || strcmp(fs->separator[n]," ") != 0)
            fprintf(stderr, fs->separator[n]);
         }
      fprintf(stderr,"\n");
      }
   }

val attrib_ptr(struct flow_info *fip, int a)
{
   val v;

   switch (a) {
   case FTFLOWINDEX:
      v.intval = &fip->FlowIndex;
      break;
   case FTLOWINTERFACE:
      v.intval = &fip->LowInterface;
      break;
   case FTLOWADJACENTTYPE:
      v.intval = &fip->LowAdjType;
      break;
   case FTLOWADJACENTADDRESS:
      v.charval = fip->LowAdjAddress;
      break;
   case FTLOWADJACENTMASK:
      v.charval = fip->LowAdjMask;
      break;
   case FTLOWPEERTYPE:
      v.intval = &fip->LowPeerType;
      break;
   case FTLOWPEERADDRESS:
      v.charval = fip->LowPeerAddress;
      break;
   case FTLOWPEERMASK:
      v.charval = fip->LowPeerMask;
      break;
   case FTLOWTRANSTYPE:
      v.intval = &fip->LowTransType;
      break;
   case FTLOWTRANSADDRESS:
      v.intval = &fip->LowTransAddress;
      break;
   case FTLOWTRANSMASK:
      v.intval = &fip->LowTransMask;
      break;
   case FTHIINTERFACE:
      v.intval = &fip->HighInterface;
      break;
   case FTHIADJACENTTYPE:
      v.intval = &fip->HighAdjType;
      break;
   case FTHIADJACENTADDRESS:
      v.charval = fip->HighAdjAddress;
      break;
   case FTHIADJACENTMASK:
      v.charval = fip->HighAdjMask;
      break;
   case FTHIPEERTYPE:
      v.intval = &fip->HighPeerType;
      break;
   case FTHIPEERADDRESS:
      v.charval = fip->HighPeerAddress;
      break;
   case FTHIPEERMASK:
      v.charval = fip->HighPeerMask;
      break;
   case FTHITRANSTYPE:
      v.intval = &fip->HighTransType;
      break;
   case FTHITRANSADDRESS:
      v.intval = &fip->HighTransAddress;
      break;
   case FTHITRANSMASK:
      v.intval = &fip->HighTransMask;
      break;
   case FTRULESET:
      v.intval = &fip->FlowRuleSet;
      break;
   case FTUPOCTETS:
      v.c64val = &fip->FwdBytes;
      break;
   case FTUPPDUS:
      v.c64val = &fip->FwdPackets;
      break;
   case FTDOWNOCTETS:
      v.c64val = &fip->BackBytes;
      break;
   case FTDOWNPDUS:
      v.c64val = &fip->BackPackets;
      break;
   case FTFIRSTTIME:
      v.intval = &fip->FirstTime;
      break;
   case FTLASTTIME:
      v.intval = &fip->LastTime;
      break;

   case FTLOWSUBSCRIBERID:
      v.charval = fip->UserID;
      break;
   case FTHISUBSCRIBERID:
      v.intval = &fip->UserCoS;
      break;
   case FTSESSIONID:
      v.intval = &fip->UserInfo;
      break;

   case FTSOURCECLASS:
      v.intval = &fip->LowClass;
      break;
   case FTDESTCLASS:
      v.intval = &fip->HighClass;
      break;
   case FTFLOWCLASS:
      v.intval = &fip->FlowClass;
      break;
   case FTSOURCEKIND:
      v.intval = &fip->LowKind;
      break;
   case FTDESTKIND:
      v.intval = &fip->HighKind;
      break;
   case FTFLOWKIND:
      v.intval = &fip->FlowKind;
      break;

   case FTAPPDNSTLD:
      v.intval = &fip->DNStld;
      break;
   case FTAPPP2PKIND:
      v.intval = &fip->p2pKind;
      break;
   case FTAPPATR4:
      v.intval = &fip->AppAtr4;
      break;
   case FTAPPATR3:
      v.intval = &fip->AppAtr3;
      break;
   case FTAPPATR2:
      v.intval = &fip->AppAtr2;
      break;
   case FTAPPATR1:
      v.intval = &fip->AppAtr1;
      break;

   case FTSPARE1:
      v.intval = &fip->Spare1;
      break;
   case FTUDPSTREAMS:
      v.intval = &fip->UDPStreams;
      break;
   case FTTCPSTREAMS:
      v.intval = &fip->TCPStreams;
      break;
   case FTTWOWAYSTREAMS:
      v.intval = &fip->TwoWayStreams;
      break;
   case FTTOTALSTREAMS:
      v.intval = &fip->TotalStreams;
      break;

   case FTPAIRINTERFACE:
      v.intval = &fip->PairInterface;
      break;
   case FTOBSINTERFACE:
      v.intval = &fip->ObsInterface;
      break;
   case FTPRIORITY:
      v.intval = &fip->Priority;
      break;
   case FTDSCODEPOINT:
      v.intval = &fip->DSCodePoint;
      break;
   case FTECNCODEBITS:
      v.intval = &fip->ECNCodeBits;
      break;
   case FTVLANID:
      v.intval = &fip->VLANid;
      break;
   case FTUPOCTETS_D:
      v.c64val = &fip->upbci;
      break;
   case FTUPPDUS_D:
      v.c64val = &fip->uppci;
      break;
   case FTDOWNOCTETS_D:
      v.c64val = &fip->dnbci;
      break;
   case FTDOWNPDUS_D:
      v.c64val = &fip->dnpci;
      break;
   case FTTAGNBR:
      v.intval = &fip->TagNbr;
      break;

   case FTUDPSTREAMS_D:
      v.intval = &fip->d_UDPStreams;
      break;
   case FTTCPSTREAMS_D:
      v.intval = &fip->d_TCPStreams;
      break;
   case FTTWOWAYSTREAMS_D:
      v.intval = &fip->d_TwoWayStreams;
      break;
   case FTTOTALSTREAMS_D:
      v.intval = &fip->d_TotalStreams;
      break;

#if defined(NETFLOW)
   case FTMETERID:
      v.intval = &fip->MeterId;
      break;
   case FTLOWROUTEASN:
      v.intval = &fip->LowRouteASN;
      break;
   case FTLOWROUTEPREFIX:
      v.intval = &fip->LowRoutePrefix;
      break;
   case FTHIROUTEASN:
      v.intval = &fip->HighRouteASN;
      break;
   case FTHIROUTEPREFIX:
      v.intval = &fip->HighRoutePrefix;
      break;
#endif

#if NEW_ATR
   case FTTCPSTREAMDATA:
      v.tsval = &fip->tsd;
      break;
   case FTTCPSTREAMDATA_D:
      v.tsval = &fip->tsd_rate;
      break;

   case FTTCPDATA:
      v.sfval = &fip->sfd;
      break;
   case FTTCPDATA_D:
      v.sfval = &fip->sfd_rate;
      break;

   case FTTOLOSTPDUS:
      v.intval = &fip->ToLostPDUs;
      break;
   case FTFROMLOSTPDUS:
      v.intval = &fip->FromLostPDUs;
      break;
   case FTTOLOSTPDUS_D:
      v.intval = &fip->d_ToLostPDUs;
      break;
   case FTFROMLOSTPDUS_D:
      v.intval = &fip->d_FromLostPDUs;
      break;
   case FTTOPQOVERFLOWS:
      v.intval = &fip->ToPQOverflows;
      break;
   case FTFROMPQOVERFLOWS:
      v.intval = &fip->FromPQOverflows;
      break;
   case FTTOPQOVERFLOWS_D:
      v.intval = &fip->d_ToPQOverflows;
      break;
   case FTFROMPQOVERFLOWS_D:
      v.intval = &fip->d_FromPQOverflows;
      break;

   case FTTOPACKETSIZE:
   case FTFROMPACKETSIZE:
   case FTTOINTERARRIVALTIME:
   case FTFROMINTERARRIVALTIME:
   case FTTOTURNAROUNDTIME:
   case FTFROMTURNAROUNDTIME:
   case FTTOBITRATE:
   case FTFROMBITRATE:
   case FTTOPDURATE:
   case FTFROMPDURATE:

   case FTTOTCPSIZE:
   case FTFROMTCPSIZE:
   case FTTOTCPTIME:
   case FTFROMTCPTIME:
   case FTTOTCPRATE1:
   case FTFROMTCPRATE1:
   case FTTOTCPRATE2:
   case FTFROMTCPRATE2:

     /*   case FTTOTURNAROUNDTIME4:
	  case FTFROMTURNAROUNDTIME4: */
   case FTTOBYTEVPKTDENSITY:
   case FTFROMBYTEVPKTDENSITY:

   case FTTOBYTEDENSITY:
   case FTFROMBYTEDENSITY:
   case FTTOPKTDENSITY:
   case FTFROMPKTDENSITY:
   case FTTOSTREAMS:
   case FTFROMSTREAMS:

   case FTTOFLOWOCTETS:
   case FTFROMFLOWOCTETS:
   case FTTOFLOWPDUS:
   case FTFROMFLOWPDUS:
   case FTFLOWTIME:

   case FTSPAREDISTRIB:

   case FTLTMINSTREAMPDUS:
   case FTLTMINSTREAMOCTETS:
   case FTTODNSRECSIZE:
   case FTFROMDNSRECSIZE:
      v.distval = &fip->dist;  /* Start of flow's distrib chain */
      break;

   case FTTOPACKETSIZE_D:
   case FTFROMPACKETSIZE_D:
   case FTTOINTERARRIVALTIME_D:
   case FTFROMINTERARRIVALTIME_D:
   case FTTOTURNAROUNDTIME_D:
   case FTFROMTURNAROUNDTIME_D:
   case FTTOBITRATE_D:
   case FTFROMBITRATE_D:
   case FTTOPDURATE_D:
   case FTFROMPDURATE_D:

   case FTTOTCPSIZE_D:
   case FTFROMTCPSIZE_D:
   case FTTOTCPTIME_D:
   case FTFROMTCPTIME_D:
   case FTTOTCPRATE1_D:
   case FTFROMTCPRATE1_D:
   case FTTOTCPRATE2_D:
   case FTFROMTCPRATE2_D:

     /*   case FTTOTURNAROUNDTIME4_D:
	  case FTFROMTURNAROUNDTIME4_D: */
   case FTTOBYTEVPKTDENSITY_D:
   case FTFROMBYTEVPKTDENSITY_D:

   case FTTOBYTEDENSITY_D:
   case FTFROMBYTEDENSITY_D:
   case FTTOPKTDENSITY_D:
   case FTFROMPKTDENSITY_D:
   case FTTOSTREAMS_D:
   case FTFROMSTREAMS_D:

   case FTTOFLOWOCTETS_D:
   case FTFROMFLOWOCTETS_D:
   case FTTOFLOWPDUS_D:
   case FTFROMFLOWPDUS_D:
   case FTFLOWTIME_D:

   case FTSPAREDISTRIB_D:

   case FTLTMINSTREAMPDUS_D:
   case FTLTMINSTREAMOCTETS_D:
   case FTTODNSRECSIZE_D:
   case FTFROMDNSRECSIZE_D:
      v.distval = &fip->dist_rate;
      break;
#endif  /* NEW_ATR */

   case FTH_HOSTS:
      v.intval = &fip->Ext2;
      break;
   case FTH_STREAMS1:
      v.intval = &fip->Ext3;
      break;
   case FTH_STREAMS2:
      v.intval = &fip->Ext4;
      break;
   case FTH_STREAMS3:
      v.intval = &fip->Ext5;
      break;

   case FTH_HOSTS_D:
      v.intval = &fip->d_Ext2;
      break;
   case FTH_STREAMS1_D:
      v.intval = &fip->d_Ext3;
      break;
   case FTH_STREAMS2_D:
      v.intval = &fip->d_Ext4;
      break;
   case FTH_STREAMS3_D:
      v.intval = &fip->d_Ext5;
      break;
   case FTMINSTREAMPDUS:
      v.intval = &fip->minstreamPDUs;
      break;

   default:
      v.intval = NULL;
      break;
      }
   return v;
   }

/* get_value() and put_value() are simplified routines for working
   with flow data files.  Strings are delimited by blanks! */
char *get_value(val v, int a, char *pb)
{
   unsigned long val;
   int j;
   unsigned char *p;
#if NEW_ATR
   struct distribution *d, *ld;
   int nb;
#endif  /* NEW_ATR */
#if VALUE_TRACE
   char buf[30];
#endif

#if TESTING
   fprintf(stderr, "    gv(1): a=%d\n", a);
   fflush(stderr);
#endif
   if (*pb == '\0') return NULL;
   while (isspace(*pb)) pb++;
#if VALUE_TRACE
   for (j = 0; j != 10; ++j) fprintf(stderr, "%c", pb[j]);
   fprintf(stderr, "  ");
#endif
   switch (a) {  /* counter64 attributes */
   case FTUPOCTETS:
   case FTUPPDUS:
   case FTDOWNOCTETS:
   case FTDOWNPDUS:
   case FTUPOCTETS_D:
   case FTUPPDUS_D:
   case FTDOWNOCTETS_D:
   case FTDOWNPDUS_D:
      pb = getc64(v.c64val, pb);
#if VALUE_TRACE
      putc64(buf, v.c64val);
      fprintf(stderr, "gv(%d) => %s\n", a, buf);
#endif
      break;

#if NEW_ATR
   case FTTCPSTREAMDATA:
   case FTTCPSTREAMDATA_D:
      sscanf(pb, "%u", &v.tsval->ToSynRTT);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      if (!isdigit(*pb)) {  /* Found a separator! */
         v.tsval->valid = 0;
         break;  /* No samples for this TCP stream */
         }

      v.tsval->valid = 1;
      sscanf(pb, "%u", &v.tsval->FromSynRTT);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &v.tsval->ToDataPoints);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &v.tsval->FromDataPoints);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.tsval->ToTotalRTT, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.tsval->FromTotalRTT, pb);
      while (isspace(*pb)) ++pb;
      break;

   case FTTCPDATA:
   case FTTCPDATA_D:
      tcpdata_in_format = 1;
      sscanf(pb, "%u", &v.sfval->n_subflows);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      if (v.sfval->n_subflows == 0)
         break;  /* No TCPdata values */

      sscanf(pb, "%u", &v.sfval->mx_active_subflows);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->ToTCPLenOctets, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->FromTCPLenOctets, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->ToTCPSeqOctets, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->FromTCPSeqOctets, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->ToTCPAckOctets, pb);
      while (isspace(*pb)) ++pb;
      pb = getc64(&v.sfval->FromTCPAckOctets, pb);
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &v.sfval->ToTCPDecrSeq);
      while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &v.sfval->FromTCPDecrSeq);
      while (isdigit(*pb)) ++pb;
      break;

   case FTTOPACKETSIZE:
   case FTFROMPACKETSIZE:
   case FTTOINTERARRIVALTIME:
   case FTFROMINTERARRIVALTIME:
   case FTTOTURNAROUNDTIME:
   case FTFROMTURNAROUNDTIME:
   case FTTOBITRATE:
   case FTFROMBITRATE:
   case FTTOPDURATE:
   case FTFROMPDURATE:

   case FTTOTCPSIZE:
   case FTFROMTCPSIZE:
   case FTTOTCPTIME:
   case FTFROMTCPTIME:
   case FTTOTCPRATE1:
   case FTFROMTCPRATE1:
   case FTTOTCPRATE2:
   case FTFROMTCPRATE2:

     /*   case FTTOTURNAROUNDTIME4:
	  case FTFROMTURNAROUNDTIME4: */
   case FTTOBYTEVPKTDENSITY:
   case FTFROMBYTEVPKTDENSITY:

   case FTTOBYTEDENSITY:
   case FTFROMBYTEDENSITY:
   case FTTOPKTDENSITY:
   case FTFROMPKTDENSITY:
   case FTTOSTREAMS:
   case FTFROMSTREAMS:

   case FTTOFLOWOCTETS:
   case FTFROMFLOWOCTETS:
   case FTTOFLOWPDUS:
   case FTFROMFLOWPDUS:
   case FTFLOWTIME:

   case FTSPAREDISTRIB:

   case FTLTMINSTREAMPDUS:
   case FTLTMINSTREAMOCTETS:
   case FTTODNSRECSIZE:
   case FTFROMDNSRECSIZE:

   case FTTOPACKETSIZE_D:
   case FTFROMPACKETSIZE_D:
   case FTTOINTERARRIVALTIME_D:
   case FTFROMINTERARRIVALTIME_D:
   case FTTOTURNAROUNDTIME_D:
   case FTFROMTURNAROUNDTIME_D:
   case FTTOBITRATE_D:
   case FTFROMBITRATE_D:
   case FTTOPDURATE_D:
   case FTFROMPDURATE_D:

   case FTTOTCPSIZE_D:
   case FTFROMTCPSIZE_D:
   case FTTOTCPTIME_D:
   case FTFROMTCPTIME_D:
   case FTTOTCPRATE1_D:
   case FTFROMTCPRATE1_D:
   case FTTOTCPRATE2_D:
   case FTFROMTCPRATE2_D:

     /*   case FTTOTURNAROUNDTIME4_D:
	  case FTFROMTURNAROUNDTIME4_D: */
   case FTTOBYTEVPKTDENSITY_D:
   case FTFROMBYTEVPKTDENSITY_D:

   case FTTOBYTEDENSITY_D:
   case FTFROMBYTEDENSITY_D:
   case FTTOPKTDENSITY_D:
   case FTFROMPKTDENSITY_D:
   case FTTOSTREAMS_D:
   case FTFROMSTREAMS_D:

   case FTTOFLOWOCTETS_D:
   case FTFROMFLOWOCTETS_D:
   case FTTOFLOWPDUS_D:
   case FTFROMFLOWPDUS_D:
   case FTFLOWTIME_D:

   case FTSPAREDISTRIB_D:

   case FTLTMINSTREAMPDUS_D:
   case FTLTMINSTREAMOCTETS_D:
   case FTTODNSRECSIZE_D:
   case FTFROMDNSRECSIZE_D:
      if (*v.distval == NULL) {
         ld = d = NULL;
         }
      else for (d = *v.distval; ; ) {  /* Find end of chain */
         if (d->selector == a) break;
         ld = d;  d = d->next;
         if (d == NULL) break;
         }
      if (d == NULL) {
         if ((d = get_dist()) == NULL) {
            fprintf(stderr, "Couldn't get space for distribution!\n");
            exit(12);
            }
#if TEST_DELTA
         fprintf(stderr, "get_val(0x%0lx, %d) -> 0x%0lx\n", v.distval,a, d);
#endif
         memset(d, 0, sizeof(struct distribution));
         if (ld == NULL) *v.distval = d;  /* First dist in chain */
         else ld->next = d;
         d->next = NULL;  d->selector = a;  /* Attribute number */
         }

      sscanf(pb, "%u", &d->Transform);  while (isdigit(*pb)) ++pb;
#if TEST_TRANSFORM
      fprintf(stderr, "--- attrib=%d, d->Transform=%d\n", a, d->Transform);
#endif
      if (d->Transform == 0) {
         /* quack(1234);  /* */
         break;  /* No data available */
         }
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->ScaleFactor);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->LowerLimit);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->UpperLimit);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->Buckets);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->Parameter1);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->Parameter2);  while (isdigit(*pb)) ++pb;
      while (isspace(*pb)) ++pb;
      sscanf(pb, "%u", &d->Parameter3);  while (isdigit(*pb)) ++pb;
#if TEST_TRANSFORM
      fprintf(stderr, "   %u %u %u %u  %u %u %u %u\n",
	      d->Transform, d->ScaleFactor, d->LowerLimit,d->UpperLimit,
	      d->Buckets, d->Parameter1, d->Parameter2, d->Parameter3);
#endif
      for (j = 0; j != d->Buckets; ++j) {
         while (isspace(*pb)) ++pb;
         pb = getc64(&d->counts[j], pb);
         }
      if (d->Transform != DS_DYN_PTS) {  /* Get overflow bin */
         while (isspace(*pb)) ++pb;
         pb = getc64(&d->counts[d->Buckets], pb);
         }
      break;
#endif  /* NEW_ATR */

   default:
      if (string_value(a)) {
         for (j = 0, p = v.charval; 
               *pb && *pb != ' ' && *pb != '\t' && *pb != ';' && j < STR_LEN;
               ++j)
            *p++ = *pb++;
         *p = '\0';
         }
      else {  /* Unsigned int */
         sscanf(pb, "%u", v.intval);
         while (isdigit(*pb)) ++pb;
#if VALUE_TRACE
         fprintf(stderr, "gv(%d) => %u\n", a, *v.intval);
#endif
         }
      }
   return pb;
   }

void put_value(int a, val v)
   /* Changed to build up record in outbuf[], 23 Dec 04 */
{
   char buf[50];
   double r;
#if NEW_ATR
   struct distribution *d;
   int j, t;
#endif  /* NEW_ATR */

   switch (a) {  /* counter64 attributes */
   case FTUPOCTETS:
   case FTUPPDUS:
   case FTDOWNOCTETS:
   case FTDOWNPDUS:
   case FTUPOCTETS_D:
   case FTUPPDUS_D:
   case FTDOWNOCTETS_D:
   case FTDOWNPDUS_D:
      if (real_out) {
         r = double64p(v.c64val);
         sprintf(buf,"%g ",r);
         }
      else putc64(buf, v.c64val);
      obp = strmov(obp, buf);
      break;

#if NEW_ATR
   case FTTCPSTREAMDATA:
   case FTTCPSTREAMDATA_D:
      if (!v.tsval->valid) {
         obp = strmov(obp,"0");  /* No samples for this TCP stream */
         break;
         }
      sprintf(buf,"%u %u %u %u ",
         v.tsval->ToSynRTT, v.tsval->FromSynRTT,
         v.tsval->ToDataPoints, v.tsval->FromDataPoints);
      obp = strmov(obp, buf);
      putc64(buf, &v.tsval->ToTotalRTT);
      obp = strmov(obp, buf);
      putc64(buf, &v.tsval->FromTotalRTT);
      obp = strmov(obp, buf);
      break;

   case FTTCPDATA:
   case FTTCPDATA_D:
      if (v.sfval->n_subflows == 0) {  /* No TCPdata values */
         obp = strmov(obp, "0");
         break;
         }
      sprintf(buf,"%u %u ",
         v.sfval->n_subflows, v.sfval->mx_active_subflows);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->ToTCPLenOctets);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->FromTCPLenOctets);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->ToTCPSeqOctets);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->FromTCPSeqOctets);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->ToTCPAckOctets);
      obp = strmov(obp, buf);
      putc64(buf, &v.sfval->FromTCPAckOctets);
      obp = strmov(obp, buf);
      sprintf(buf,"%u %u",
         v.sfval->ToTCPDecrSeq, v.sfval->FromTCPDecrSeq);
      obp = strmov(obp, buf);
      break;

   case FTTOPACKETSIZE:
   case FTFROMPACKETSIZE:
   case FTTOINTERARRIVALTIME:
   case FTFROMINTERARRIVALTIME:
   case FTTOTURNAROUNDTIME:
   case FTFROMTURNAROUNDTIME:
   case FTTOBITRATE:
   case FTFROMBITRATE:
   case FTTOPDURATE:
   case FTFROMPDURATE:

   case FTTOTCPSIZE:
   case FTFROMTCPSIZE:
   case FTTOTCPTIME:
   case FTFROMTCPTIME:
   case FTTOTCPRATE1:
   case FTFROMTCPRATE1:
   case FTTOTCPRATE2:
   case FTFROMTCPRATE2:

     /*   case FTTOTURNAROUNDTIME4:
	  case FTFROMTURNAROUNDTIME4: */
   case FTTOBYTEVPKTDENSITY:
   case FTFROMBYTEVPKTDENSITY:

   case FTTOBYTEDENSITY:
   case FTFROMBYTEDENSITY:
   case FTTOPKTDENSITY:
   case FTFROMPKTDENSITY:
   case FTTOSTREAMS:
   case FTFROMSTREAMS:

   case FTTOFLOWOCTETS:
   case FTFROMFLOWOCTETS:
   case FTTOFLOWPDUS:
   case FTFROMFLOWPDUS:
   case FTFLOWTIME:

   case FTSPAREDISTRIB:

   case FTLTMINSTREAMPDUS:
   case FTLTMINSTREAMOCTETS:
   case FTTODNSRECSIZE:
   case FTFROMDNSRECSIZE:

   case FTTOPACKETSIZE_D:
   case FTFROMPACKETSIZE_D:
   case FTTOINTERARRIVALTIME_D:
   case FTFROMINTERARRIVALTIME_D:
   case FTTOTURNAROUNDTIME_D:
   case FTFROMTURNAROUNDTIME_D:
   case FTTOBITRATE_D:
   case FTFROMBITRATE_D:
   case FTTOPDURATE_D:
   case FTFROMPDURATE_D:

   case FTTOTCPSIZE_D:
   case FTFROMTCPSIZE_D:
   case FTTOTCPTIME_D:
   case FTFROMTCPTIME_D:
   case FTTOTCPRATE1_D:
   case FTFROMTCPRATE1_D:
   case FTTOTCPRATE2_D:
   case FTFROMTCPRATE2_D:

     /*   case FTTOTURNAROUNDTIME4_D:
	  case FTFROMTURNAROUNDTIME4_D: */
   case FTTOBYTEVPKTDENSITY_D:
   case FTFROMBYTEVPKTDENSITY_D:

   case FTTOBYTEDENSITY_D:
   case FTFROMBYTEDENSITY_D:
   case FTTOPKTDENSITY_D:
   case FTFROMPKTDENSITY_D:
   case FTTOSTREAMS_D:
   case FTFROMSTREAMS_D:

   case FTTOFLOWOCTETS_D:
   case FTFROMFLOWOCTETS_D:
   case FTTOFLOWPDUS_D:
   case FTFROMFLOWPDUS_D:
   case FTFLOWTIME_D:

   case FTSPAREDISTRIB_D:

   case FTLTMINSTREAMPDUS_D:
   case FTLTMINSTREAMOCTETS_D:
   case FTTODNSRECSIZE_D:
   case FTFROMDNSRECSIZE_D:
      for (d = *v.distval; d != NULL; d = d->next) {
         if (d->selector == a) break;
         }
      if (d == NULL) {
         fprintf(stderr, "pv(): Couldn't find distribution attrib %d\n", a);
         exit(12);
         }

      t = d->Transform;
      if (t == DS_LIN_RESET) t = DS_LIN;
      else if (t == DS_LOG_RESET) t = DS_LOG;
      sprintf(buf,"%u",t);
      obp = strmov(obp, buf);
      if (d->Transform == 0) break;  /* No data available */

      sprintf(buf," %u %u %u  %u %u %u %u ",
         d->ScaleFactor, d->LowerLimit, d->UpperLimit,
         d->Buckets, d->Parameter1, d->Parameter2, d->Parameter3);
      obp = strmov(obp, buf);
      for (j = 0; j != d->Buckets; ++j) {
         if (real_out) {
            r = double64p(&d->counts[j]);
            sprintf(buf,"%g",r);
            }
         else putc64(buf, &d->counts[j]);
         *obp++ = ' ';
         obp = strmov(obp, buf);
         }
      if (d->Transform != DS_DYN_PTS) {
         if (real_out) {
            r = (double)d->counts[j];  /* Oflo bucket */
            sprintf(buf,"%g",r);
            }
         else putc64(buf, &d->counts[j]);  /* Oflo bucket */
         *obp++ = ' ';
         obp = strmov(obp, buf);
         }
      break;
#endif  /* NEW_ATR */

   default:
      if (string_value(a)) {
         sprintf(buf,"%s",v.charval);
         }
      else {
         sprintf(buf,"%u",*v.intval);
	 }
      obp = strmov(obp, buf);
      }

   *obp = '\0';
   }



char *strmov(char *d, char *s)
{
   while (*s != '\0') *d++ = *s++;
   return d;
   }

static char ftb[32];
static char *moy[] = {
   "Jan", "Feb", "Mar", "Apr", "May", "Jun",
   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

char *fmt_time(time_t *ti, int utc)
{
   if (!utc) {
      char *ts = ctime(ti);
      strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
      strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
      strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
      strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
      strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
      ftb[24] = '\0';
      }
   else {
      struct tm *t = gmtime(ti);
      char *dow[] = {
         "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
      sprintf(ftb, "%02d:%02d:%02d %s %2d %s %4d",
         t->tm_hour,t->tm_min,t->tm_sec, dow[t->tm_wday],
         t->tm_mday, moy[t->tm_mon], t->tm_year+1900);
      }
   return ftb;
   }   

static char stb[32];

char *str_time(char *fts)  /* Convert .fdf time to yyyymmdd.HHMMSS string */
{
  int h,m,s,d,y, j;

   sscanf(fts, "%2u:%2u:%2u %*s %2u %s %4u", &h,&m,&s, &d,stb,&y);
   for (j = 0; j != 12; ++j) {
      if (strncmp(stb, moy[j], 3) == 0) break;
      }
   sprintf(stb, "%04u%02u%02u.%02u%02u%02u", y,j+1,d, h,m,s);
   return stb;
   }
