/* 1245, Mon 17 May 99

   RTFM_ATR.H:  RTFM attribute definitions

   Copyright (C) 1998-2006 by Nevil Brownlee,
   CAIDA | University of Auckland */

/*
 * $Log: rtfm_atr.h,v $
 * Revision 1.1.1.2  2006/03/21 18:08:55  nevil
 * Version 5.1b15
 *
 * Revision 1.1.1.2.2.19  2002/10/31 23:02:40  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * Revision 1.1.1.2.2.18  2002/09/13 15:05:33  nevil
 * Implement VLANid, Priority and ECNCodeBits attributes
 *
 * Revision 1.1.1.2.2.17  2002/08/27 17:13:21  nevil
 * Improve thread locking.  Implement STREAM_TO_FLOW, ECNBits and TCPStreamData
 *
 * Revision 1.1.1.2.2.16  2002/07/03 01:45:57  nevil
 * Implement OAM attributes
 *
 * Revision 1.1.1.2.2.15  2002/06/30 01:07:07  nevil
 * Allow define text > 1000 chars, allow *Mask attribs in save and format
 *
 * Revision 1.1.1.2.2.14  2002/06/29 01:56:08  nevil
 * Miscellaneous fixes for warnings found by Compaq Unix cc
 *
 * Revision 1.1.1.2.2.7  2000/11/29 21:40:17  nevil
 * Bug fixes for stream / packet queue handling
 *
 * Revision 1.1.1.2.2.6  2000/08/08 19:44:59  nevil
 * 44b8 release
 *
 * Revision 1.1.1.2.2.4  2000/06/06 03:38:32  nevil
 * Combine NEW_ATR with TCP_ATR, various bug fixes
 *
 * Revision 1.1.1.2.2.1  2000/01/12 02:57:15  nevil
 * Implement 'packet pair matched' turnaroundtime distribution attributes.
 * Fix ASN-related bugs in NeTraMet, distribution-related bugs in fd_filter.
 *
 * Revision 1.1.1.2  1999/10/03 21:06:33  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.7  1999/05/18 03:36:30  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.6  1999/04/09 03:49:00  nevil
 * Implemented IPv6 addresses for SRL
 *   -D define V6 to enable the v6 code
 *   Added IPv4 and IPv6 as address families (IP => IPv4)
 *   Added get_v6address ro srl_scan
 *   v6err() routine to print v6 address errors
 *   V6DEBUG turns on v6 parser debugging
 *
 * Revision 1.1.1.1.2.5  1999/01/28 03:12:09  nevil
 * Mis-spelt attribute names are now correctly reported as errors
 *
 * Revision 1.1.1.1.2.4  1999/01/27 04:26:16  nevil
 * Minor corrections to fix compiler warnings
 *
 * Revision 1.1.1.1.2.3  1999/01/08 01:38:40  nevil
 * Distribution file for 4.3b7
 *
 * Revision 1.1.1.1.2.2  1998/11/19 02:13:25  nevil
 * Allow mask_from_width() to accept zero width
 *
 * Revision 1.1.1.1.2.1  1998/11/17 23:02:42  nevil
 * Implement DS attributes SourceDScodepoint(118), DestDScodepoint(119)
 * and (experimental) TCP attribute TCPdata(121)
 *
 * Revision 1.1.1.1  1998/11/16 03:22:03  nevil
 * Import of release 4.3b3
 *
 * Revision 1.1.2.1  1998/10/22 21:40:36  nevil
 * Moved srl from src/manager to its own subdirectory
 *
 * Revision 1.1.3.2  1998/10/18 23:44:12  nevil
 * Added Nicolai's patches, some 'tidying up' of the source
 *
 * Revision 1.1.3.1  1998/10/13 02:48:26  nevil
 * Import of Nicolai's 4.2.2
 *
 * Revision 1.1.1.1  1998/08/24 12:09:29  nguba
 * NetraMet 4.2 Original Distribution
 *
 * Revision 1.2  1998/07/21 00:43:57  rtfm
 * Change attrib numbers for 'New Attribs' I-D
 * First release version of SRL
 */

#ifndef RTFMEXT
#define RTFMEXT  extern
#define DECLARE  0
#define INIT(v)
#else
#define RTFMEXT
#define DECLARE  1
#define INIT(v)  = v
#endif


#define MAC_ADDR_LEN      6  /* Address lengths */

#define IP4_ADDR_LEN      4
#define IP6_ADDR_LEN     16
#define DN_ADDR_LEN       3
#define ET_ADDR_LEN       3
#if FULL_IPX
# define IPX_ADDR_LEN    10
# undef 
#else
# define IPX_ADDR_LEN     4
#endif
#define NSAP_ADDR_LEN    20

#define TRANS_ADDR_LEN    2

#if CLNS
# define PEER_ADDR_LEN  NSAP_ADDR_LEN
# define RULE_ADDR_LEN  NSAP_ADDR_LEN
#elif V6 || FULL_IPX
# define PEER_ADDR_LEN   IP6_ADDR_LEN
# define RULE_ADDR_LEN   IP6_ADDR_LEN
#else  /* v4 */
# define PEER_ADDR_LEN   IP4_ADDR_LEN
# define RULE_ADDR_LEN   MAC_ADDR_LEN
#endif

#define SESSION_ID_LEN  128

RTFMEXT char addr_len[]  /* Lengths for each address type */
#if DECLARE
   = {
      PEER_ADDR_LEN,    /*  0 */
      IP4_ADDR_LEN,     /*  1 IP4 */
      IP6_ADDR_LEN,     /*  2 IP6 */
      NSAP_ADDR_LEN,    /*  3 CLNS */
      0,                /*  4 */
      0,                /*  5 */
      TRANS_ADDR_LEN,   /*  6 Other */
      0,                /*  7 */
      0,                /*  8 */
      0,                /*  9 */
      0,                /* 10 */
      IPX_ADDR_LEN,     /* 11 Novell IPX */
      ET_ADDR_LEN,      /* 12 EtherTalk */
      DN_ADDR_LEN       /* 13 DECnet */
      }
#endif
   ;


#define AT_DUMMY       0  /* Addr_type values */

#define AT_IP4         1  /* Address Family Numbers */
#define AT_IP6         2
#define AT_CLNS        3
#define AT_OTHER       6  /* 'generic' 802 address family */
#define AT_ETHERNET    7  /* ifType medium */
#define AT_TOKENRING   9  /* ifType medium */
#define AT_NOVELL     11
#define AT_ETHERTALK  12
#define AT_DECNET     13
#define AT_FDDI       15  /* ifType medium */
#define AT_AAL5       49  /* ifType AAL5 over ATM */

#define MX_PROTOCOLS  AT_DECNET

#define AT3_IGNORE     0  /* Addr_type values */
#define AT3_DUMMY      1
#define AT3_ADJACENT   1
#define AT3_IP         2
#define AT3_CLNS       3
#define AT3_IDPR       4
#define AT3_DECNET     5
#define AT3_NOVELL     6
#define AT3_ETHERTALK  7
#define AT3_TRANS      8
#define AT3_IP_FRAG   11
#define AT3_OTHER     12

#define PT_ICMP       1  /* IP protocol type values */
#define PT_TCP        6
#define PT_UDP       17
#define PT_OSPF      89  /* OSPF IGP */

#define WNP_FTPDATA  20  /* Well-known tcp/udp port numbers */
#define WNP_FTP      21
#define WNP_TELNET   23
#define WNP_SMTP     25
#define WNP_DOMAIN   53
#define WNP_GOPHER   70
#define WNP_WWW      80
#define WNP_NNTP    119
#define WNP_NTP     123
#define WNP_SNMP    161

#define RA_SUCCEED       0  /* Ignore */
#define RA_IGNORE        1
#define RA_NOMATCH       2
#define RA_COUNT         3
#define RA_COUNTPKT      4
#define RA_RETURN        5
#define RA_GOSUB         6
#define RA_GOSUBACT      7
#define RA_ASSIGN        8
#define RA_ASSIGNACT     9
#define RA_GOTO         10
#define RA_GOTOACT      11
#define RA_PUSHTO       12
#define RA_PUSHTOACT    13
#define RA_PUSHPKTTO    14
#define RA_PUSHPKTTOACT 15
#define RA_POPTO        16
#define RA_POPTOACT     17

#define RA_NEXT         21  /* 'the next rule,' not an action */

#define FTFLOWINDEX		1  /* Flow table attribute values */
#define FTFLOWTIMEMARK          2
#define FTFLOWSTATUS		3

#define FTLOWINTERFACE          4
#define FTLOWADJACENTTYPE       5
#define FTLOWADJACENTADDRESS    6
#define FTLOWADJACENTMASK       7
#define FTLOWPEERTYPE		8
#define FTLOWPEERADDRESS	9
#define FTLOWPEERMASK	       10
#define FTLOWTRANSTYPE         11
#define FTLOWTRANSADDRESS      12
#define FTLOWTRANSMASK         13

#define FTHIINTERFACE          14
#define FTHIADJACENTTYPE       15
#define FTHIADJACENTADDRESS    16
#define FTHIADJACENTMASK       17
#define FTHIPEERTYPE	       18
#define FTHIPEERADDRESS	       19
#define FTHIPEERMASK	       20
#define FTHITRANSTYPE          21
#define FTHITRANSADDRESS       22
#define FTHITRANSMASK          23

#define FTPDUSCALE             24
#define FTOCTETSCALE           25
#define FTRULESET              26
#define FTUPOCTETS	       27
#define FTUPPDUS	       28
#define FTDOWNOCTETS	       29
#define FTDOWNPDUS	       30
#define FTFIRSTTIME	       31
#define FTLASTTIME	       32
#define FTLASTACTIVETIME       32

#define FTLOWSUBSCRIBERID      33
#define FTHISUBSCRIBERID       34
#define FTSESSIONID            35

#define FTSOURCECLASS          36
#define FTDESTCLASS            37
#define FTFLOWCLASS            38
#define FTSOURCEKIND           39
#define FTDESTKIND             40
#define FTFLOWKIND             41

        /* Room for more 'new basic' attributes below 109
              LASTDISTRIB attribute nbr is 97 */

#define FTAPPDNSTLD            98
#define FTAPPP2PKIND           99
#define FTAPPATR4             100
#define FTAPPATR3             101
#define FTAPPATR2             102
#define FTAPPATR1             103
#define N_APP_ATTRS             6

#define FTSPARE1              104
#define FTUDPSTREAMS          105
#define FTTCPSTREAMS          106
#define FTTWOWAYSTREAMS       107
#define FTTOTALSTREAMS        108

#define FTPAIRINTERFACE       109
#define FTOBSINTERFACE        110
#define FTTCPSTREAMDATA       111  /* <TCPStreamData> */

#define FTPRIORITY            117
#define FTDSCODEPOINT         118
#define FTECNCODEBITS         119
#define FTVLANID              120

#define N_NEW_BASIC  12+N_APP_ATTRS  /* New 'basic' attributes */

#if !NEW_ATR
#define N_OLD_ATRS    (FTFLOWKIND + N_NEW_BASIC)
#define LASTATTRIB       FTVLANID   /* May be redefined below */
#else
#define FTTOLOSTPDUS          121
#define FTFROMLOSTPDUS        122
#define FTTOPQOVERFLOWS       123  /* PQOF */
#define FTFROMPQOVERFLOWS     124
#define N_OLD_ATRS     (FTFLOWKIND + N_NEW_BASIC + 4)  /* + OverflowPDUs */
#define LASTATTRIB      FTFROMPQOVERFLOWS    /* May be redefined below */
#endif

#define FTFORWARD              50
#define FTV1                   51
#define FTV2                   52
#define FTV3                   53
#define FTV4                   54
#define FTV5                   55

#define USER_ATTRIB(a)  (a >= FTV1 && a <= FTV5)
#define N_METER_ATRS    (FTV5+1 - FTFORWARD)
#define N_VBLS          (FTV5+1 - FTV1)  /* Nbr of user variables */

#if !NEW_ATR
#define N_DIST_ATRS        0
#else
#define FTDISTRIBUTIONS        65  /* Bit set for each distribution in flow */

#define FTTOPACKETSIZE         66  /* Attributes 66..75 defined in RFC 2724 */
#define FTFROMPACKETSIZE       67
#define FTTOINTERARRIVALTIME   68
#define FTFROMINTERARRIVALTIME 69
#define FTTOTURNAROUNDTIME     70
#define FTFROMTURNAROUNDTIME   71
#define FTTOBITRATE            72
#define FTFROMBITRATE          73
#define FTTOPDURATE            74
#define FTFROMPDURATE          75

#define FTTOTCPTIME            76
#define FTFROMTCPTIME          77
#define FTTOTCPSIZE            78
#define FTFROMTCPSIZE          79
#define FTTOTCPRATE1           80
#define FTFROMTCPRATE1         81
#define FTTOTCPRATE2           82
#define FTFROMTCPRATE2         83

/* #define FTTOTURNAROUNDTIME4    88
   #define FTFROMTURNAROUNDTIME4  89 */
#define FTTOBYTEVPKTDENSITY    88
#define FTFROMBYTEVPKTDENSITY  89

#define FTTOBYTEDENSITY        84
#define FTFROMBYTEDENSITY      85
#define FTTOPKTDENSITY         86
#define FTFROMPKTDENSITY       87

#define FTTOFLOWOCTETS         90
#define FTFROMFLOWOCTETS       91
#define FTTOFLOWPDUS           92
#define FTFROMFLOWPDUS         93
#define FTFLOWTIME             94

#define FTTOSTREAMS            95
#define FTFROMSTREAMS          96

#define FTSPAREDISTRIB         97

#define LAST_DISTRIB           97  /* 65+32 = 97 */

#define DISTRIB_BASE       FTTOPACKETSIZE
#define N_DIST_ATRS   (1 + (LAST_DISTRIB-FTDISTRIBUTIONS + 2))
   /* +2 for synonyms, To/FromTurnaroundTime1 */
# if LASTATTRIB < LAST_DISTRIB
#  undef LASTATTRIB
#  define LASTATTRIB      LAST_DISTRIB  /* max(basic, dist) */
# endif
#endif

#if !NETFLOW
#define N_NF_ATRS               0
#else
#define FTMETERID             112
#define FTLOWROUTEASN         113
#define FTLOWROUTEPREFIX      114
#define FTHIROUTEASN          115
#define FTHIROUTEPREFIX       116

#define NF_ATTRIB(a)  (a >= FTMETERID && a <= FTHIASWIDTH)
#define N_NF_ATRS     (FTHIROUTEPREFIX+1 - FTMETERID)
# if LASTATTRIB < FTHIROUTEPREFIX
#  undef LASTATTRIB
#  define LASTATTRIB  FTHIROUTEPREFIX  /* max(basic, dist, nf) */
# endif
#endif

#define FTTCPDATA             125
#define FTOAMIDENT            126
#define FTOAMDATA             127
#define N_TCP_ATRS              3
# if LASTATTRIB < FTOAMDATA
#  undef LASTATTRIB
#  define LASTATTRIB    FTOAMDATA  /* max(basic, dist, nf, tcp) */
# endif

#define EXT_ATR_BASE          128
#define FTHOSTDATA            129
#define FTH_HOSTS             130
#define FTH_STREAMS1          131
#define FTH_STREAMS2          132
#define FTH_STREAMS3          133
#define FTMINSTREAMPDUS       134

/* Make room for another set of distributions ... */
#define FTDISTRIBS2           135 /* Bit set for each distrib2 in flow */
#define FTLTMINSTREAMPDUS     FTDISTRIBS2+1
#define FTLTMINSTREAMOCTETS   FTDISTRIBS2+2
#define FTTODNSRECSIZE        FTDISTRIBS2+3 
#define FTFROMDNSRECSIZE      FTDISTRIBS2+4 

#define LAST_DISTRIB2         FTDISTRIBS2+4  /* Max is +32 */
#define LAST_EXT_ATR          LAST_DISTRIB2

#define DISTRIB2_ATTRIB(a)   (a >= FTLTMINSTREAMPDUS && a <= LAST_DISTRIB2)
#define DISTRIB2_BASE        FTLTMINSTREAMPDUS
#define N_DIST2_ATRS         (1 + LAST_DISTRIB2-FTDISTRIBS2)

# if LASTATTRIB < LAST_EXT_ATR
#  undef LASTATTRIB
#  define LASTATTRIB    LAST_EXT_ATR  /* max(basic, dist, nf, tcp, ext) */
# endif
#define N_EXT_ATRS  (LAST_EXT_ATR - EXT_ATR_BASE)

#define N_ATTRIBS  (N_OLD_ATRS + N_METER_ATRS + N_DIST_ATRS \
                      + N_NF_ATRS + N_TCP_ATRS + N_EXT_ATRS)

#define DISTRIB_ATTRIB(a) ( (a >= FTTOPACKETSIZE    && a <= LAST_DISTRIB) || \
                            (a >= FTLTMINSTREAMPDUS && a <=  LAST_DISTRIB2) )

#if NEW_ATR
#define DIST_PARAM_LEN  6
#define MXBUCKETS     100
struct distribution {
   unsigned char selector,  /* Attribute number */
      mask_params[DIST_PARAM_LEN], value_params[DIST_PARAM_LEN];
   unsigned int counts[MXBUCKETS+1];

   unsigned char Transform, ScaleFactor; /* 'mask' parameters */
   unsigned int LowerLimit, UpperLimit;
   unsigned char Buckets, Parameter1;  /* 'value' parameters */
   unsigned int Parameter2, Parameter3;

   struct distribution *next;
   };
#define DS_NULL  0  /* Distribution transform type values */
#define DS_LIN   1
#define DS_LOG   2
#endif


#define ADJ_ADDR    1  /* Address attribute types */
#define PEER_ADDR   2
#define TRANS_ADDR  3


struct attrib_info {
   char *name;
   unsigned char index;
   unsigned char properties;
   unsigned char len;
   };

RTFMEXT int attr_ix[LASTATTRIB+1];  /* Index to attribs[] */

#define if_OK(x)      ((attribs[attr_ix[x]].properties & 0x08) != 0)
#define save_OK(x)    ((attribs[attr_ix[x]].properties & 0x04) != 0)
#define store_OK(x)   ((attribs[attr_ix[x]].properties & 0x02) != 0)
#define format_OK(x)  ((attribs[attr_ix[x]].properties & 0x01) != 0)

#define P(ifbit, savebit, storebit, formatbit)  \
   (ifbit*0x08 | savebit*0x04 | storebit*0x02 | formatbit)

#define N_SYNONYMS   1
/* Synonyms: LastActiveTime */
#define SZ_ATTRIBS  (1 + N_ATTRIBS + N_SYNONYMS)  /* 1 for Null */

RTFMEXT struct attrib_info attribs[SZ_ATTRIBS]
#if DECLARE
   = {
   /*  name                 index             (if save store format)  len */
   "null",                  0,                      P(0,0,0,0), 0,
   "flowindex",             FTFLOWINDEX,            P(0,0,0,1), 4,
   "flowtimemark",          FTFLOWTIMEMARK,         P(0,0,0,0), 4,
   "flowstatus",            FTFLOWSTATUS,           P(0,0,0,0), 1,
   "sourceinterface",       FTLOWINTERFACE,         P(1,1,0,1), TRANS_ADDR_LEN,
   "sourceadjacenttype",    FTLOWADJACENTTYPE,      P(1,1,0,1), 1,
   "sourceadjacentaddress", FTLOWADJACENTADDRESS,   P(1,1,0,1), MAC_ADDR_LEN,
   "sourceadjacentmask",    FTLOWADJACENTMASK,      P(0,1,0,1), MAC_ADDR_LEN,
   "sourcepeertype",        FTLOWPEERTYPE,          P(1,1,0,1), 1,
   "sourcepeeraddress",     FTLOWPEERADDRESS,       P(1,1,0,1), PEER_ADDR_LEN,
   "sourcepeermask",        FTLOWPEERMASK,          P(0,1,0,1), PEER_ADDR_LEN,
   "sourcetranstype",       FTLOWTRANSTYPE,         P(1,1,0,1), 1,
   "sourcetransaddress",    FTLOWTRANSADDRESS,      P(1,1,0,1), TRANS_ADDR_LEN,
   "sourcetransmask",       FTLOWTRANSMASK,         P(0,1,0,1), TRANS_ADDR_LEN,

   "destinterface",         FTHIINTERFACE,          P(1,1,0,1), TRANS_ADDR_LEN,
   "destadjacenttype",      FTHIADJACENTTYPE,       P(1,1,0,1), 1,
   "destadjacentaddress",   FTHIADJACENTADDRESS,    P(1,1,0,1), MAC_ADDR_LEN,
   "destadjacentmask",      FTHIADJACENTMASK,       P(0,1,0,1), MAC_ADDR_LEN,
   "destpeertype",          FTHIPEERTYPE,           P(1,1,0,1), 1,
   "destpeeraddress",       FTHIPEERADDRESS,        P(1,1,0,1), PEER_ADDR_LEN,
   "destpeermask",          FTHIPEERMASK,           P(0,1,0,1), PEER_ADDR_LEN,
   "desttranstype",         FTHITRANSTYPE,          P(1,1,0,1), 1,
   "desttransaddress",      FTHITRANSADDRESS,       P(1,1,0,1), TRANS_ADDR_LEN,
   "desttransmask",         FTHITRANSMASK,          P(0,1,0,1), TRANS_ADDR_LEN,

   "pduscale",              FTPDUSCALE,             P(0,0,0,0), 0,
   "octetscale",            FTOCTETSCALE,           P(0,0,0,0), 0,
   "flowruleset",           FTRULESET,              P(1,1,0,1), 1,
   "tooctets",              FTUPOCTETS,             P(0,0,0,1), 8,
   "topdus",                FTUPPDUS,               P(0,0,0,1), 8,
   "fromoctets",            FTDOWNOCTETS,           P(0,0,0,1), 8,
   "frompdus",              FTDOWNPDUS,             P(0,0,0,1), 8,
   "firsttime",             FTFIRSTTIME,            P(0,0,0,1), 4,
   "lasttime",              FTLASTTIME,             P(0,0,0,1), 4,
   "lastactivetime",        FTLASTTIME,             P(0,0,0,1), 4,

   "sessionid",             FTLOWSUBSCRIBERID,      P(0,1,0,1), SESSION_ID_LEN,
   "usercos",               FTHISUBSCRIBERID,       P(1,1,0,1), 1,
   "userinfo",              FTSESSIONID,            P(1,1,0,1), 1,

   "sourceclass",           FTSOURCECLASS,          P(1,0,1,1), 1,
   "destclass",             FTDESTCLASS,            P(1,0,1,1), 1,
   "flowclass",             FTFLOWCLASS,            P(1,0,1,1), 1,
   "sourcekind",            FTSOURCEKIND,           P(1,0,1,1), 1,
   "destkind",              FTDESTKIND,             P(1,0,1,1), 1,
   "flowkind",              FTFLOWKIND,             P(1,0,1,1), 1

#if NEW_ATR
   ,
   "tcpstreamdata",         FTTCPSTREAMDATA,        P(0,1,0,1), 1
#endif                         /* Mask and value not used for TCP data */
   ,
   "dnstld",                FTAPPDNSTLD,            P(0,1,0,1), TRANS_ADDR_LEN,
   "p2pkind",               FTAPPP2PKIND,           P(0,1,0,1), TRANS_ADDR_LEN,
   "apppatr4",              FTAPPATR4,              P(0,1,0,1), TRANS_ADDR_LEN,
   "apppatr3",              FTAPPATR3,              P(0,1,0,1), TRANS_ADDR_LEN,
   "apppatr2",              FTAPPATR2,              P(0,1,0,1), TRANS_ADDR_LEN,
   "apppatr1",              FTAPPATR1,              P(0,1,0,1), TRANS_ADDR_LEN,

   "spare1",                FTSPARE1,               P(0,1,0,1), TRANS_ADDR_LEN,
   "udpstreams",            FTUDPSTREAMS,           P(0,1,0,1), TRANS_ADDR_LEN,
   "tcpstreams",            FTTCPSTREAMS,           P(0,1,0,1), TRANS_ADDR_LEN,
   "twowaystreams",         FTTWOWAYSTREAMS,        P(0,1,0,1), TRANS_ADDR_LEN,
   "totalstreams",          FTTOTALSTREAMS,         P(0,1,0,1), TRANS_ADDR_LEN,

   "pairinterface",         FTPAIRINTERFACE,        P(1,1,0,1), TRANS_ADDR_LEN,
                       /* Cached-matching  synonym for SourceInterface */
   "obsinterface",          FTOBSINTERFACE,         P(1,1,0,1), 1,
   "priority",              FTPRIORITY,             P(1,1,0,1), 1,
   "dscodepoint",           FTDSCODEPOINT,          P(1,1,0,1), 1,
   "ecncodebits",           FTECNCODEBITS,          P(1,1,0,1), 1,
   "vlanid",                FTVLANID,               P(1,1,0,1), TRANS_ADDR_LEN,

   "tolostpdus",            FTTOLOSTPDUS,           P(0,1,0,1), 1,
   "fromlostpdus",          FTFROMLOSTPDUS,         P(0,1,0,1), 1,
   "topqoverflows",         FTTOPQOVERFLOWS,        P(0,1,0,1), 1,
   "frompqoverflows",       FTFROMPQOVERFLOWS,      P(0,1,0,1), 1,

   "matchingstod",          FTFORWARD,              P(1,0,0,0), 1,
   "v1",                    FTV1,                   P(0,0,0,0), RULE_ADDR_LEN,
   "v2",                    FTV2,                   P(0,0,0,0), RULE_ADDR_LEN,
   "v3",                    FTV3,                   P(0,0,0,0), RULE_ADDR_LEN,
   "v4",                    FTV4,                   P(0,0,0,0), RULE_ADDR_LEN,
   "v5",                    FTV5,                   P(0,0,0,0), RULE_ADDR_LEN

#if NEW_ATR
   ,
   "distributions",         FTDISTRIBUTIONS,        P(0,0,0,0), 0,
   "topacketsize",          FTTOPACKETSIZE,         P(0,1,0,1), DIST_PARAM_LEN,
   "frompacketsize",        FTFROMPACKETSIZE,       P(0,1,0,1), DIST_PARAM_LEN,
   "tointerarrivaltime",    FTTOINTERARRIVALTIME,   P(0,1,0,1), DIST_PARAM_LEN,
   "frominterarrivaltime",  FTFROMINTERARRIVALTIME, P(0,1,0,1), DIST_PARAM_LEN,
   "toturnaroundtime1",     FTTOTURNAROUNDTIME,     P(0,1,0,1), DIST_PARAM_LEN,
   "fromturnaroundtime1",   FTFROMTURNAROUNDTIME,   P(0,1,0,1), DIST_PARAM_LEN,
   "toturnaroundtime",      FTTOTURNAROUNDTIME,     P(0,1,0,1), DIST_PARAM_LEN,
   "fromturnaroundtime",    FTFROMTURNAROUNDTIME,   P(0,1,0,1), DIST_PARAM_LEN,
   "tobitrate",             FTTOBITRATE,            P(0,1,0,1), DIST_PARAM_LEN,
   "frombitrate",           FTFROMBITRATE,          P(0,1,0,1), DIST_PARAM_LEN,
   "topktrate",             FTTOPDURATE,            P(0,1,0,1), DIST_PARAM_LEN,
   "frompktrate",           FTFROMPDURATE,          P(0,1,0,1), DIST_PARAM_LEN,
   "totcptime",             FTTOTCPTIME,            P(0,1,0,1), DIST_PARAM_LEN,
   "fromtcptime",           FTFROMTCPTIME,          P(0,1,0,1), DIST_PARAM_LEN,
   "totcpsize",             FTTOTCPSIZE,            P(0,1,0,1), DIST_PARAM_LEN,
   "fromtcpsize",           FTFROMTCPSIZE,          P(0,1,0,1), DIST_PARAM_LEN,
   "totcprate1",            FTTOTCPRATE1,           P(0,1,0,1), DIST_PARAM_LEN,
   "fromtcprate1",          FTFROMTCPRATE1,         P(0,1,0,1), DIST_PARAM_LEN,
   "totcprate2",            FTTOTCPRATE2,           P(0,1,0,1), DIST_PARAM_LEN,
   "fromtcprate2",          FTFROMTCPRATE2,         P(0,1,0,1), DIST_PARAM_LEN,

   /*   "toturnaroundtime4",     FTTOTURNAROUNDTIME4,    P(0,1,0,1), DIST_PARAM_LEN,
	"fromturnaroundtime4",   FTFROMTURNAROUNDTIME4,  P(0,1,0,1), DIST_PARAM_LEN, */
   "tobytevpktdensity",     FTTOBYTEVPKTDENSITY,    P(0,1,0,1), DIST_PARAM_LEN,
   "frombytevpktdensity",   FTFROMBYTEVPKTDENSITY,  P(0,1,0,1), DIST_PARAM_LEN,

   "tobytedensity",         FTTOBYTEDENSITY,        P(0,1,0,1), DIST_PARAM_LEN,
   "frombytedensity",       FTFROMBYTEDENSITY,      P(0,1,0,1), DIST_PARAM_LEN,
   "topktdensity",          FTTOPKTDENSITY,         P(0,1,0,1), DIST_PARAM_LEN,
   "frompktdensity",        FTFROMPKTDENSITY,       P(0,1,0,1), DIST_PARAM_LEN,
   "tostreams",             FTTOSTREAMS,            P(0,1,0,1), DIST_PARAM_LEN,
   "fromstreams",           FTFROMSTREAMS,          P(0,1,0,1), DIST_PARAM_LEN,

   "toflowoctets",          FTTOFLOWOCTETS,         P(0,1,0,1), DIST_PARAM_LEN,
   "fromflowoctets",        FTFROMFLOWOCTETS,       P(0,1,0,1), DIST_PARAM_LEN,
   "toflowpdus",            FTTOFLOWPDUS,           P(0,1,0,1), DIST_PARAM_LEN,
   "fromflowpdus",          FTFROMFLOWPDUS,         P(0,1,0,1), DIST_PARAM_LEN,
   "flowtime",              FTFLOWTIME,             P(0,1,0,1), DIST_PARAM_LEN,

   "sparedistrib",          FTSPAREDISTRIB,         P(0,1,0,1), DIST_PARAM_LEN
#endif

#if NETFLOW
   ,
   "meterid",               FTMETERID,              P(1,1,0,1), 1,
   "sourceasn",             FTLOWROUTEASN,          P(1,1,0,1), TRANS_ADDR_LEN,
   "sourceprefix",          FTLOWROUTEPREFIX,       P(1,1,0,1), 1,
   "destasn",               FTHIROUTEASN,           P(1,1,0,1), TRANS_ADDR_LEN,
   "destprefix",            FTHIROUTEPREFIX,        P(1,1,0,1), 1
#endif

#if NEW_ATR
   ,
   "tcpdata",               FTTCPDATA,              P(0,1,0,1), 1,
                                /* Mask and value not used for TCP data */
   "oamident",              FTOAMIDENT,             P(0,1,1,1), 1,
   "oamdata",               FTOAMDATA,              P(0,1,0,1), DIST_PARAM_LEN
#endif
   ,
   "hostdata",              FTHOSTDATA,             P(0,1,0,1), TRANS_ADDR_LEN,
   "h_hosts",               FTH_HOSTS,              P(0,1,0,1), TRANS_ADDR_LEN,
   "h_streams1",            FTH_STREAMS1,           P(0,1,0,1), TRANS_ADDR_LEN,
   "h_streams2",            FTH_STREAMS2,           P(0,1,0,1), TRANS_ADDR_LEN,
   "h_streams3",            FTH_STREAMS3,           P(0,1,0,1), TRANS_ADDR_LEN,
   "minstreampdus",         FTMINSTREAMPDUS,        P(0,0,0,1), 1,

   "distributions2",        FTDISTRIBS2,            P(0,0,0,0), 0,
   "ltminstreampdus",       FTLTMINSTREAMPDUS,      P(0,0,0,1), DIST_PARAM_LEN,
   "ltminstreamoctets",     FTLTMINSTREAMOCTETS,    P(0,0,0,1), DIST_PARAM_LEN,
   "todnsrecordsize",       FTTODNSRECSIZE,         P(0,1,0,1), DIST_PARAM_LEN,
   "fromdnsrecordsize",     FTFROMDNSRECSIZE,       P(0,1,0,1), DIST_PARAM_LEN,
   }
#endif
   ;

#undef DECLARE
#undef INIT
